/*
 * FileDisk Shell Extension
 * Copyright (c) 2005-2006 Gianluigi Tiesi <sherpya@netfarm.it>
 * ShellExt Code taken from ClamWin Antivirus Scanner (c) 2004 alch,
 * FileDisk Code taken from FileDisk (c) 1999-2004 Bo Brantn.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "FileDiskExt.h"

void PrintLastError(LPTSTR title, DWORD flags)
{
    LPTSTR lpMsgBuf;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        0,
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
        );

    MessageBox(NULL, (LPTSTR) lpMsgBuf, title, MB_OK | flags);
    LocalFree(lpMsgBuf);
}

void PrintLastError(PTCHAR title)
{
#ifdef _DEBUG
    PrintLastError(title, 0);
#endif
}

/*
    FUNCTION: CShellExt::Initialize(LPCITEMIDLIST, LPDATAOBJECT, HKEY)

    PURPOSE: Called by the shell when initializing a context menu or property
             sheet extension.

    PARAMETERS:
        pIDFolder - Specifies the parent folder
        pDataObj  - Spefifies the set of items selected in that folder.
        hRegKey   - Specifies the type of the focused item in the selection.

    RETURN VALUE:

        NOERROR in all cases.

    COMMENTS:   Note that at the time this function is called, we don't know
                (or care) what type of shell extension is being initialized.
                It could be a context menu or a property sheet.
*/

STDMETHODIMP CShellExt::Initialize(LPCITEMIDLIST pIDFolder, LPDATAOBJECT pDataObj, HKEY hRegKey)
{
    HRESULT hres = E_FAIL;
    STGMEDIUM medium;
    FORMATETC fmte = {CF_HDROP, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};


    /* Initialize can be called more than once */
    if (m_pDataObj)
        m_pDataObj->Release();

    /* duplicate the object pointer and registry handle */
    if (pDataObj)
    {
        m_pDataObj = pDataObj;
        pDataObj->AddRef();
    }

    /* use the given IDataObject to get a list of filenames (CF_HDROP) */
    hres = pDataObj->GetData(&fmte, &medium);

    if(FAILED(hres)) return E_FAIL;

    m_szPath[0] = 0;
    DragQueryFile((HDROP) medium.hGlobal, 0, m_szPath, sizeof(m_szPath));
    ::ReleaseStgMedium(&medium);

    return NOERROR;
}

BOOL CShellExt::IsMounted(LPTSTR szPath)
{
    size_t len = 0, i = 0;
    int type = 0;
    TCHAR szDrive[MAX_PATH];

    len = GetLogicalDriveStrings(MAX_PATH, szDrive);
    while (i < len)
    {
        type = GetDriveType(&szDrive[i]);
        if ((type != DRIVE_REMOTE) && (type != DRIVE_REMOVABLE))
        {
            memset(m_szMountedImage, 0, sizeof(m_szMountedImage));
            if (!FileDiskStatus(*(&szDrive[i])) && !_tcsicmp(m_szMountedImage, szPath))
            {
                m_MountedOn = PathGetDriveNumber(&szDrive[i]);
                return TRUE;
            }
        }
        i += ((_tcslen(&szDrive[i]) + 1) * sizeof(TCHAR));
    }

    return FALSE;
}

UINT CShellExt::GetType(LPTSTR szPath)
{
    DWORD type = GetDriveType(szPath);

    if (PathMatchSpec(szPath, _T("*.iso")))
    {
        if (IsMounted(szPath))
            return TYPE_ISO_MOUNTED;
        else
            return TYPE_ISO_UNMOUNTED;
    }

    if (PathMatchSpec(szPath, _T("*.img")))
    {
        if (IsMounted(szPath))
            return TYPE_IMG_MOUNTED;
        else
            return TYPE_IMG_UNMOUNTED;
    }

    if ((type != DRIVE_UNKNOWN) && (type != DRIVE_REMOTE) && (type != DRIVE_REMOVABLE))
    {
        if (FileDiskStatus(*(&szPath[0])))
            return TYPE_DRIVE_NOTFILEDISK;
        else
        {
            m_MountedOn = PathGetDriveNumber(szPath);
            return TYPE_DRIVE_MOUNTED;
        }
    }

    return TYPE_NOT_HANDLED;
}

/*
    FUNCTION: CShellExt::QueryContextMenu(HMENU, UINT, UINT, UINT, UINT)

    PURPOSE: Called by the shell just before the context menu is displayed.
             This is where you add your specific menu items.

    PARAMETERS:
        hMenu      - Handle to the context menu
        indexMenu  - Index of where to begin inserting menu items
        idCmdFirst - Lowest value for new menu ID's
        idCmtLast  - Highest value for new menu ID's
        uFlags     - Specifies the context of the menu event

    RETURN VALUE:

    COMMENTS:
*/

/* The menu text */
STDMETHODIMP CShellExt::QueryContextMenu(HMENU hMenu, UINT indexMenu, UINT idCmdFirst, UINT idCmdLast, UINT uFlags)
{

    UINT        idCmd = idCmdFirst;
    HRESULT     hr = E_INVALIDARG;
    TCHAR       buf[MAX_PATH];

    m_Type = GetType(m_szPath);

    /* Seperator */
    InsertMenu(hMenu, indexMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);

    switch (m_Type)
    {
        case TYPE_DRIVE_MOUNTED:
            _stprintf(buf, _T("Umount %s"), m_szMountedImage);
            InsertMenu(hMenu, indexMenu, MF_STRING|MF_BYPOSITION, idCmd++, buf);
            if (m_hFileDiskBmp) SetMenuItemBitmaps (hMenu, indexMenu, MF_BYPOSITION, m_hFileDiskBmp, m_hFileDiskBmp);
            indexMenu++;
            break;
        case TYPE_ISO_MOUNTED:
            _stprintf(buf, _T("Umount iso image from drive %c:"), m_MountedOn+65);
            InsertMenu(hMenu, indexMenu, MF_STRING|MF_BYPOSITION, idCmd++, buf);
            if (m_hFileDiskBmp) SetMenuItemBitmaps (hMenu, indexMenu, MF_BYPOSITION, m_hFileDiskBmp, m_hFileDiskBmp);
            indexMenu++;
            break;
        case TYPE_IMG_MOUNTED:
            _stprintf(buf, _T("Umount image from drive %c:"), m_MountedOn+65);
            InsertMenu(hMenu, indexMenu, MF_STRING|MF_BYPOSITION, idCmd++, buf);
            if (m_hFileDiskBmp) SetMenuItemBitmaps (hMenu, indexMenu, MF_BYPOSITION, m_hFileDiskBmp, m_hFileDiskBmp);
            indexMenu++;
            break;
        case TYPE_ISO_UNMOUNTED:
        case TYPE_IMG_UNMOUNTED:
        {
            DWORD res = GetLogicalDrives();
            TCHAR szItem[32];
            HMENU hPopUp = CreatePopupMenu();
            idCmd++;
            InsertMenu(hMenu, indexMenu, MF_STRING|MF_BYPOSITION|MF_POPUP, (UINT_PTR) hPopUp, _T("Mount via Filedisk"));
            if (m_hFileDiskBmp) SetMenuItemBitmaps (hMenu, indexMenu, MF_BYPOSITION, m_hFileDiskBmp, m_hFileDiskBmp);
            indexMenu++;
            for (int i=0; i<26; i++)
            {
                if ((res >> i) & 0x1)
                {
                    idCmd++;
                    continue;
                }

                _stprintf(szItem, _T("on letter %c:"), 65 + i);
                InsertMenu(hPopUp, indexMenu++, MF_STRING|MF_BYPOSITION, idCmd++, szItem);
            }

        }
        break;
    default:
        return NULL;
    }

    /* Seperator */
    InsertMenu(hMenu, indexMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);

    return ResultFromShort(idCmd-idCmdFirst); /* Must return number of menu */
    /* items we added */
}

BOOL CShellExt::Umount(HWND hwnd)
{
    if (!m_NumberOfDevices)
    {
        MessageBox(hwnd, _T("FileDisk driver is not installed correctly"), _T("FileDisk"), MB_OK | MB_ICONERROR);
        return FALSE;
    }

    switch (m_Type)
    {
        case TYPE_ISO_MOUNTED:
        case TYPE_DRIVE_MOUNTED:
            if (FileDiskUmount(m_MountedOn+65))
                PrintLastError(_T("FileDisk: Error umounting device"), MB_ICONERROR);
            else
                PrintLastError(_T("FileDisk Umount"));
            break;
    }
    return TRUE;
}

BOOL CShellExt::Mount(HWND hwnd, TCHAR DriveLetter)
{
    OPEN_FILE_INFORMATION OpenFileInformation;
    DWORD i = 0;
    bool failed = true;
    TCHAR FileName[MAX_PATH];

    memset(&OpenFileInformation, 0, sizeof(OPEN_FILE_INFORMATION));
    FileName[0] = _T('\0');

    if (m_szPath[0] == _T('\\'))
    {
        if (m_szPath[1] == _T('\\'))
        /* \\server\share\path\filedisk.img */
        {
            _tcsncpy(FileName, _T("\\??\\UNC"), MAX_PATH);
            _tcsncat(FileName, m_szPath + sizeof(TCHAR), MAX_PATH);
        }
        else
        /* \Device\Harddisk0\Partition1\path\filedisk.img */
            _tcsncpy(FileName, m_szPath, MAX_PATH);
    }
    else
    /* c:\path\filedisk.img */
    {
        _tcsncpy(FileName, _T("\\??\\"), MAX_PATH);
        _tcsncat(FileName, m_szPath, MAX_PATH);
    }

#ifdef _UNICODE
    if (!WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, FileName, -1, (LPSTR) OpenFileInformation.FileName, MAX_PATH, NULL, NULL))
    {
        PrintLastError(_T("WideCharToMultiByte() Failed"), MB_ICONERROR);
        return FALSE;
    }
#else
    memcpy(OpenFileInformation.FileName, FileName, strlen(FileName) + 1);
#endif

    OpenFileInformation.FileNameLength = (USHORT) _tcslen(FileName);
    OpenFileInformation.ReadOnly = TRUE;

    /* Check if the image can be mounted read/write */
    if (PathMatchSpec(m_szPath, _T("*.img")) && !(GetFileAttributes(m_szPath) & FILE_ATTRIBUTE_READONLY))
        OpenFileInformation.ReadOnly = FALSE;

    for (i = 0; i < m_NumberOfDevices; i++)
    {
        if (!FileDiskMount(i, &OpenFileInformation, DriveLetter, TRUE))
        {
            failed = false;
            break;
        }
    }

    if (failed)
    {
        PrintLastError(_T("FileDisk: Cannot mount the device, try increasing max number of devices"), MB_ICONERROR);
        return FALSE;
    }

    PrintLastError(_T("FileDisk: Mount Iso"));

    return TRUE;
}

void CShellExt::SetImagePath(LPTSTR szPath)
{
    m_szPath[0] = 0;
    _tcsncat(m_szPath, szPath, MAX_PATH);
}

/*
    FUNCTION: CShellExt::InvokeCommand(LPCMINVOKECOMMANDINFO)

    PURPOSE: Called by the shell after the user has selected on of the
             menu items that was added in QueryContextMenu().

    PARAMETERS:
        lpcmi - Pointer to an CMINVOKECOMMANDINFO structure

    RETURN VALUE: HRESULT code signifying success or failure;NOERROR if no error

    COMMENTS:
*/

STDMETHODIMP CShellExt::InvokeCommand(LPCMINVOKECOMMANDINFO lpcmi)
{
    HRESULT hr = E_INVALIDARG;

    /* If HIWORD(lpcmi->lpVerb) then we have been called programmatically
       and lpVerb is a command that should be invoked.  Otherwise, the shell
       has called us, and LOWORD(lpcmi->lpVerb) is the menu ID the user has
       selected.  Actually, it's (menu ID - idCmdFirst) from QueryContextMenu().
    */

    if (!HIWORD(lpcmi->lpVerb))
    {
        UINT idCmd = LOWORD(lpcmi->lpVerb);
        switch (idCmd)
        {
            case 0:
                Umount(lpcmi->hwnd);
                break;
            default:
                Mount(lpcmi->hwnd, idCmd+64);
        }
        hr = NOERROR;
    }
    return hr;
}

/*
    FUNCTION: CShellExt::GetCommandString(UINT_PTR idCmd, UINT uFlags, UINT FAR *reserved, LPSTR pszName, UINT cchMax)

    PURPOSE: Called by the shell to retreive the cononical command name
             or help text for a menu item added by a context menu extension handler

    PARAMETERS:
        idCmd - Menu item ID offset
        uFlags - Value specifying the type of information to retreive
        *reserved - Pointer to reserved value
        pszName - Pointer to buffer to receive name string or help text
        cchMax - Buffer size

    RETURN VALUE: HRESULT code signifying success or failure;NOERROR if no error

    COMMENTS:
*/


STDMETHODIMP CShellExt::GetCommandString(UINT_PTR idCmd, UINT uFlags, UINT FAR *reserved, LPSTR pszName, UINT cchMax)
{
    switch (idCmd)
    {
        case 0:
            if (uFlags == GCS_HELPTEXTW)
                wcsncpy((LPWSTR) pszName, L"FileDisk Shell Extension", cchMax);
            else if (uFlags == GCS_HELPTEXTA)
                strncpy(pszName, "FileDisk Shell Extension", cchMax);
            break;
    }
    return NOERROR;
}
